# Copyright (c) 2021 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx.helper as helper

from qonnx.core.datatype import DataType
from qonnx.custom_op.base import CustomOp


def binary_quant(inp_tensor, scale):
    # ToDo: Update this link, when the PR gets merged
    # Port of IntQuant class from Brevitas: https://bit.ly/2S6qvZJ

    # Quantizing
    y_int = inp_tensor
    y_ones = np.ones(y_int.shape, dtype=y_int.dtype)
    y_int = np.where(y_int >= 0.0, y_ones, -y_ones)
    # Scaling
    out_tensor = y_int * scale

    return out_tensor


class BipolarQuant(CustomOp):
    """Bipolar quantization operation for QONNX. Takes four inputs:
    - input tensor to quantize
    - the scale

    The output is a tensor of the same shape as the input tensor, with quantized
    values.
    """

    def get_nodeattr_types(self):
        return dict()

    def make_shape_compatible_op(self, model):
        node = self.onnx_node
        return helper.make_node("Identity", [node.input[0]], [node.output[0]])

    def get_integer_datatype(self, model):
        return DataType["BIPOLAR"]

    def get_output_dtype(self, model):
        node = self.onnx_node
        # scale must be read from initializers
        scale = model.get_initializer(node.input[1])
        # determine the QONNX DataType
        unit_scale = np.all(scale == 1.0)
        if unit_scale:
            finn_dt = self.get_integer_datatype(model)
        else:
            finn_dt = DataType["FLOAT32"]

        return finn_dt

    def infer_node_datatype(self, model):
        try:
            finn_dt = self.get_output_dtype(model)
        except AssertionError:
            finn_dt = DataType["FLOAT32"]
        node = self.onnx_node
        model.set_tensor_datatype(node.output[0], finn_dt)

    def execute_node(self, context, graph):
        node = self.onnx_node
        # save inputs
        inp_tensor = context[node.input[0]]
        scale = context[node.input[1]]
        # calculate output
        ret = binary_quant(inp_tensor, scale)
        # ensure output is ndarray (even if 0d)
        # since numpy silently flattens 0d arrays to scalars
        # more: https://github.com/numpy/numpy/issues/13105
        if not isinstance(ret, np.ndarray):
            ret = np.asarray(ret, dtype=np.float32)
        # set context according to output name
        context[node.output[0]] = ret

    def verify_node(self):
        pass



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np

import qonnx.core.onnx_exec as oxe
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op import channels_last
from qonnx.custom_op.channels_last.base_wrapped_op import to_channels_last_args
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.channels_last import (
    AbsorbChanFirstIntoMatMul,
    InsertChannelsLastDomainsAndTrafos,
    MoveChanFirstDownstream,
    MoveChanLastUpstream,
    MoveTransposePastFork,
    RemoveConsecutiveChanFirstAndChanLastTrafos,
)
from qonnx.transformation.general import GiveUniqueNodeNames
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.make_input_chanlast import MakeInputChannelsLast
from qonnx.transformation.quant_constant_folding import FoldTransposeIntoQuantInit
from qonnx.util.basic import is_finn_op
from qonnx.util.test import download_model, get_golden_in_and_output, test_model_details
from qonnx.util.to_channels_last import to_channels_last

model_details_chanlast = {
    "FINN-CNV_W2A2": {
        "layout_sensitive": True,
        "final_transpose_count": 0,
    },
    "FINN-TFC_W2A2": {
        "layout_sensitive": False,
        "final_transpose_count": 0,
    },
    "RadioML_VGG10": {
        "layout_sensitive": True,
        "final_transpose_count": 0,
    },
    "Conv_bias_example": {
        "layout_sensitive": True,
        "final_transpose_count": 0,
    },
    "rn18_w4a4_a2q_16b": {
        "layout_sensitive": True,
        # TODO: RN18 has AveragePool, Trunc, Reshape ops at the end so one Tranpose
        # still gets left in there after channels-last conversion. To get rid of this final one:
        # * AveragePool needs channels-last op variant
        # * Trunc can be covered by eltwise ops. afterwards
        # * Gemm needs to be convert to MatMul with GemmToMatMul
        # * Transpose->Reshape->MatMul will be handled by AbsorbChanFirstIntoMatMul
        "final_transpose_count": 1,
    },
}
# inherit basics for matching testcases from test util
model_details = {k: v for (k, v) in test_model_details.items() if k in model_details_chanlast.keys()}
model_details = {**model_details, **model_details_chanlast}


def analysis_testing_for_chanlast_domain(model):
    # Define for each ChannelsLast operation the number of minimum dimensions, it needs to have
    ChanLast_node_types_and_min_dim_input = {
        "Conv": 3,
        "MaxPool": 3,
        "BatchNormalization": 3,
    }
    # Check that all wrapped_ops in the registry have a definition here
    chanlast_op_types = list(channels_last.custom_op.keys())
    testable_op_types = list(ChanLast_node_types_and_min_dim_input.keys())
    for op_name in chanlast_op_types:
        assert (
            op_name in testable_op_types
        ), f"The channelsLast op {op_name} is missing a definition for the domain string test."

    for n_type, min_dim in ChanLast_node_types_and_min_dim_input.items():
        nodes = model.get_nodes_by_op_type(n_type)
        for n in nodes:
            input_shape = model.get_tensor_shape(n.input[0])
            if len(input_shape) >= min_dim:
                assert (
                    n.domain == "qonnx.custom_op.channels_last"
                ), f"Node domain is not set correctly for node with name: {n.name}"
    return dict()


def analysis_test_for_left_transposes(model, test_model, make_input_channels_last):
    t_nodes = model.get_nodes_by_op_type("Transpose")
    if test_model == "FINN-TFC_W2A2":
        # For the TFC the assertion should be the other way around.
        make_input_channels_last = not make_input_channels_last
    final_transpose_count = model_details[test_model]["final_transpose_count"]
    if make_input_channels_last:
        assert len(t_nodes) == final_transpose_count, f"Unexpected final Transpose count: {len(t_nodes)}"
    else:
        assert len(t_nodes) == final_transpose_count + 1, f"Unexpected final Transpose count: {len(t_nodes)}"
    return dict()


def verify_all_nodes(model):
    result = dict()
    for n in model.graph.node:
        if is_finn_op(n.domain):
            n_instance = getCustomOp(n)
            verify_result = n_instance.verify_node()
            result[n.name] = verify_result
    return result


def analysis_first_node_is_transpose(model):
    result = dict()
    input_tensor = model.graph.input[0].name
    first_node = model.find_consumer(input_tensor)
    assert first_node.op_type == "Transpose", "First node in the network should be a transpose node."
    return result


@pytest.mark.parametrize("make_input_channels_last", [True, False])
@pytest.mark.parametrize("test_model", model_details.keys())
def test_channelslast_conversion_end2end(test_model, make_input_channels_last):
    # Download an clean model
    onnx_file = download_model(test_model, do_cleanup=True)
    input_tensor, golden_result = get_golden_in_and_output(test_model)

    # Execute transformation
    qonnx_all_trafos = onnx_file.split(".onnx")[0] + "_all_nhwc_trafos_test.onnx"
    to_channels_last(onnx_file, make_input_channels_last=make_input_channels_last, out_file=qonnx_all_trafos)

    # Check output
    model = ModelWrapper(qonnx_all_trafos)
    if make_input_channels_last:
        input_dims = len(model.get_tensor_shape(model.graph.input[0].name))
        input_tensor = input_tensor.transpose(to_channels_last_args(input_dims))
    input_dict = {model.graph.input[0].name: input_tensor}
    output_dict = oxe.execute_onnx(model, input_dict)
    current_result = output_dict[model.graph.output[0].name]
    assert np.isclose(
        golden_result, current_result, atol=1e-7
    ).all(), "Output of cleaned QONNX model and channels last model should match."
    assert model.check_all_tensor_shapes_specified(), "All tensor shapes should be specified."

    # Check that the ops, which should be ChannelsLast actually are
    _ = model.analysis(analysis_testing_for_chanlast_domain)

    # This would throw an error if anything is misconfigured
    _ = model.analysis(verify_all_nodes)

    # Check that the transposes are gone
    _ = analysis_test_for_left_transposes(model, test_model, make_input_channels_last)

    # Check that the first node is a transpose node for layout-sensitive models
    if (not make_input_channels_last) and (model_details[test_model]["layout_sensitive"]):
        _ = model.analysis(analysis_first_node_is_transpose)


@pytest.mark.parametrize("test_model", model_details.keys())
def test_channelslast_conversion_step_by_step(test_model):
    # Download an clean model
    onnx_file = download_model(test_model, do_cleanup=True)
    input_tensor, golden_result = get_golden_in_and_output(test_model)

    # Execute transformation
    model = ModelWrapper(onnx_file)
    qonnx_all_trafos = onnx_file.split(".onnx")[0] + "_all_nhwc_trafos.onnx"

    # Run trafo
    model = model.transform(InsertChannelsLastDomainsAndTrafos())
    model = model.transform(GiveUniqueNodeNames())
    # Check output
    input_dict = {model.graph.input[0].name: input_tensor}
    output_dict = oxe.execute_onnx(model, input_dict)
    current_result = output_dict[model.graph.output[0].name]
    assert np.isclose(
        golden_result, current_result, atol=1e-7
    ).all(), "Output of cleaned QONNX model and model after applying InsertChannelsLastDomainsAndTrafos should match."
    assert model.check_all_tensor_shapes_specified(), "All tensor shapes should be specified."

    # Run trafo
    model = model.transform(RemoveConsecutiveChanFirstAndChanLastTrafos())
    # Check output
    input_dict = {model.graph.input[0].name: input_tensor}
    output_dict = oxe.execute_onnx(model, input_dict)
    current_result = output_dict[model.graph.output[0].name]
    assert np.isclose(
        golden_result, current_result, atol=1e-7
    ).all(), (
        "Output of cleaned QONNX model and model after applying RemoveConsecutiveChanFirstAndChanLastTrafos should match."
    )
    assert model.check_all_tensor_shapes_specified(), "All tensor shapes should be specified."

    # Run trafo
    model = model.transform(MoveChanLastUpstream())
    # Check output
    input_dict = {model.graph.input[0].name: input_tensor}
    output_dict = oxe.execute_onnx(model, input_dict)
    current_result = output_dict[model.graph.output[0].name]
    assert np.isclose(
        golden_result, current_result, atol=1e-7
    ).all(), "Output of cleaned QONNX model and model after applying MoveChanLastUpstream should match."
    assert model.check_all_tensor_shapes_specified(), "All tensor shapes should be specified."

    # Run trafo
    model = model.transform(FoldTransposeIntoQuantInit())
    # Check output
    input_dict = {model.graph.input[0].name: input_tensor}
    output_dict = oxe.execute_onnx(model, input_dict)
    current_result = output_dict[model.graph.output[0].name]
    assert np.isclose(
        golden_result, current_result, atol=1e-7
    ).all(), "Output of cleaned QONNX model and model after applying MoveChanLastUpstream should match."
    assert model.check_all_tensor_shapes_specified(), "All tensor shapes should be specified."

    # Run trafo
    model = model.transform(RemoveConsecutiveChanFirstAndChanLastTrafos())
    # Check output
    input_dict = {model.graph.input[0].name: input_tensor}
    output_dict = oxe.execute_onnx(model, input_dict)
    current_result = output_dict[model.graph.output[0].name]
    assert np.isclose(
        golden_result, current_result, atol=1e-7
    ).all(), (
        "Output of cleaned QONNX model and model after applying RemoveConsecutiveChanFirstAndChanLastTrafos should match."
    )
    assert model.check_all_tensor_shapes_specified(), "All tensor shapes should be specified."

    # Run trafo
    model = model.transform(MoveChanFirstDownstream())
    model = model.transform(MoveTransposePastFork())
    # Check output
    input_dict = {model.graph.input[0].name: input_tensor}
    output_dict = oxe.execute_onnx(model, input_dict)
    current_result = output_dict[model.graph.output[0].name]
    assert np.isclose(
        golden_result, current_result, atol=1e-7
    ).all(), "Output of cleaned QONNX model and model after applying MoveChanFirstDownstream should match."
    assert model.check_all_tensor_shapes_specified(), "All tensor shapes should be specified."

    # Run trafo
    model = model.transform(AbsorbChanFirstIntoMatMul())
    model = model.transform(InferShapes())
    # Check output
    input_dict = {model.graph.input[0].name: input_tensor}
    output_dict = oxe.execute_onnx(model, input_dict)
    current_result = output_dict[model.graph.output[0].name]
    assert np.isclose(
        golden_result, current_result, atol=1e-7
    ).all(), "Output of cleaned QONNX model and model after applying AbsorbChanFirstIntoMatMul should match."
    assert model.check_all_tensor_shapes_specified(), "All tensor shapes should be specified."

    # Run trafo
    model = model.transform(MakeInputChannelsLast())
    model = model.transform(RemoveConsecutiveChanFirstAndChanLastTrafos())
    # Check output
    input_dims = len(model.get_tensor_shape(model.graph.input[0].name))
    input_tensor = input_tensor.transpose(to_channels_last_args(input_dims))
    input_dict = {model.graph.input[0].name: input_tensor}
    output_dict = oxe.execute_onnx(model, input_dict)
    current_result = output_dict[model.graph.output[0].name]
    assert np.isclose(
        golden_result, current_result, atol=1e-7
    ).all(), "Output of cleaned QONNX model and model after applying AbsorbChanFirstIntoMatMul should match."
    assert model.check_all_tensor_shapes_specified(), "All tensor shapes should be specified."

    model.save(qonnx_all_trafos)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import warnings

import qonnx.core.data_layout as DataLayout
import qonnx.custom_op.registry as registry
from qonnx.transformation.base import Transformation
from qonnx.util.basic import get_by_name, is_finn_op


def _dims_to_layout(model, node, ndims):
    if ndims == 2:
        return DataLayout.NC
    else:
        if is_finn_op(node.domain):
            if node.op_type == "MultiThreshold" or node.op_type == "QuantAvgPool2d":
                mt_inst = registry.getCustomOp(node)
                layout = mt_inst.get_nodeattr("data_layout")
                if layout == "NHWC" and ndims == 4:
                    return DataLayout.NHWC
                elif layout == "NCHW" and ndims == 4:
                    return DataLayout.NCHW
                elif layout == "NWC" and ndims == 3:
                    return DataLayout.NWC
                elif layout == "NC" and ndims == 2:
                    return DataLayout.NC
                else:
                    return DataLayout.UNKNOWN
            else:
                if ndims == 4:
                    return DataLayout.NHWC
                elif ndims == 3:
                    return DataLayout.NWC
                elif ndims == 2:
                    return DataLayout.NC
                else:
                    return DataLayout.UNKNOWN
        else:
            # propagate input layout to output
            # TODO this won't work for concat, squeeze/unsqueeze/reshape...
            return model.get_tensor_layout(node.input[0])


def _infer_node_data_layout(model, node):
    """Infer output data layout annotation(s) for a particular node.
    Returns True if any changes were made."""
    old_layouts = list(map(lambda x: model.get_tensor_layout(x), node.output))
    try:
        if is_finn_op(node.domain):
            # try to guess based on number of output dims
            for o in node.output:
                ndims = len(model.get_tensor_shape(o))
                new_layout = _dims_to_layout(model, node, ndims)
                model.set_tensor_layout(o, new_layout)
        else:
            if node.op_type == "Transpose":
                # grab input annotation and switch it around using perm
                perm = get_by_name(node.attribute, "perm").ints
                inp_layout = model.get_tensor_layout(node.input[0])
                out_layout = [inp_layout[i] for i in perm]
                model.set_tensor_layout(node.output[0], out_layout)
            elif node.op_type == "Unsqueeze":
                inp_layout = model.get_tensor_layout(node.input[0])
                # add dummy dimension at the output
                out_layout = inp_layout + ["x"]
                model.set_tensor_layout(node.output[0], out_layout)
            elif node.op_type == "Squeeze":
                inp_layout = model.get_tensor_layout(node.input[0])
                assert inp_layout[-1] == "x"
                # remove dummy dimension
                out_layout = inp_layout[:-1]
                model.set_tensor_layout(node.output[0], out_layout)
            else:
                # try to guess based on number of output dims
                for o in node.output:
                    ndims = len(model.get_tensor_shape(o))
                    model.set_tensor_layout(o, _dims_to_layout(model, node, ndims))
    except Exception:
        for o in node.output:
            model.set_tensor_layout(o, DataLayout.UNKNOWN)

    # compare old and new output dtypes to see if anything changed
    new_layouts = list(map(lambda x: model.get_tensor_layout(x), node.output))
    graph_modified = new_layouts != old_layouts
    return graph_modified


class InferDataLayouts(Transformation):
    """Try to infer data layout annotations info for all input/intermediate/output
    tensors based on inputs and node type."""

    def apply(self, model):
        graph = model.graph
        graph_modified = False
        # first, make sure that the global input has an annotation
        # this is really hard to do in general, so we do some bad guesswork
        inp_name = graph.input[0].name
        if model.get_tensor_layout(inp_name) is None:
            inp_shape = model.get_tensor_shape(inp_name)
            if len(inp_shape) == 4:
                warnings.warn("Assuming 4D input is NCHW")
                model.set_tensor_layout(inp_name, DataLayout.NCHW)
                graph_modified = True
            elif len(inp_shape) == 3:
                warnings.warn("Assuming 3D input is NWC")
                model.set_tensor_layout(inp_name, DataLayout.NWC)
                graph_modified = True
            elif len(inp_shape) == 2:
                graph_modified = True
                warnings.warn("Assuming 2D input is NC")
                model.set_tensor_layout(inp_name, DataLayout.NC)
            else:
                raise Exception(
                    """Unknown number of dims for input, don't know
                how to annotate"""
                )
        for node in graph.node:
            graph_modified |= _infer_node_data_layout(model, node)
        return (model, graph_modified)



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
from onnx import TensorProto, helper

import qonnx.core.onnx_exec as oxe
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.general import SortGraph
from qonnx.transformation.infer_datatypes import InferDataTypes
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.remove import RemoveIdentityOps
from qonnx.util.basic import gen_finn_dt_tensor, qonnx_make_model


def insert_identity_op(model, op, as_first_node, approx, fork_after_id):
    kwargs = {}
    inp_ndims = 4 if as_first_node else 2
    if approx:
        zero_val = 0.000001
        one_val = 0.999999
    else:
        zero_val = 0.0
        one_val = 1.0
    if op in ["Add", "Sub"]:
        val = np.asarray([zero_val], dtype=np.float32)
    elif op in ["Mul", "Div"]:
        val = np.asarray([one_val], dtype=np.float32)
    elif op in ["Identity"]:
        val = None
    elif op == "Pad":
        # opset 11 and above: padding specified as input and not attribute
        val = np.asarray([0] * 2 * inp_ndims, dtype=np.int64)
    elif op == "Dropout":
        val = None
        kwargs = {"ratio": 0.0}
    else:
        return

    graph = model.graph
    if val is None:
        inplist = ["inp" if as_first_node else "div_out"]
    else:
        model.set_initializer("value", val)
        inplist = ["inp" if as_first_node else "div_out", "value"]
    identity_node = helper.make_node(op, inplist, ["ident_out"], **kwargs)
    old_2nd_node = graph.node[1]
    old_last_node = graph.node[-1]
    graph.node.append(identity_node)
    if fork_after_id:
        graph.node.append(helper.make_node("Mul", ["ident_out", "mul2"], ["mulbranch0_out"]))
        model.set_initializer("mul2", np.asarray([2.0], dtype=np.float32))
        graph.node.append(helper.make_node("Mul", ["ident_out", "mul3"], ["mulbranch1_out"]))
        model.set_initializer("mul3", np.asarray([3.0], dtype=np.float32))
        graph.node.append(helper.make_node("Add", ["mulbranch0_out", "mulbranch1_out"], ["idfork_out"]))
        subgraph_out = "idfork_out"
    else:
        subgraph_out = "ident_out"

    if as_first_node:
        old_2nd_node.input[0] = subgraph_out
    else:
        old_last_node.input[0] = subgraph_out
    model = model.transform(SortGraph())

    return model


# identity operations to be inserted
@pytest.mark.parametrize("op", ["Add", "Sub", "Mul", "Div", "Identity", "Pad", "Dropout"])
@pytest.mark.parametrize("approx", [False, True])
@pytest.mark.parametrize("as_first_node", [False, True])
@pytest.mark.parametrize("fork_before_id", [False, True])
@pytest.mark.parametrize("fork_after_id", [False, True])
def test_remove_identity_ops(op, as_first_node, approx, fork_before_id, fork_after_id):
    if approx and not (op in ["Add", "Sub", "Mul", "Div"]):
        pytest.skip(f"approx=True not relevant for {op}")
    # set up onnx model
    inp = helper.make_tensor_value_info("inp", TensorProto.FLOAT, [1, 4, 1, 1])
    mul = helper.make_tensor_value_info("mul", TensorProto.FLOAT, [])
    shape = helper.make_tensor_value_info("shape", TensorProto.FLOAT, [2])
    div = helper.make_tensor_value_info("div", TensorProto.FLOAT, [])
    matmul = helper.make_tensor_value_info("matmul", TensorProto.FLOAT, [4, 2])
    outp = helper.make_tensor_value_info("outp", TensorProto.FLOAT, [1, 2])

    mul_node = helper.make_node("Mul", ["inp", "mul"], ["mul_out"])
    reshape_node = helper.make_node("Reshape", ["mul_out", "shape"], ["reshape_out"])
    div_node = helper.make_node("Div", ["reshape_out", "div"], ["div_out"])
    matmul_node = helper.make_node("MatMul", ["div_out", "matmul"], ["outp"])

    graph = helper.make_graph(
        nodes=[mul_node, reshape_node, div_node, matmul_node],
        name="identity-graph",
        inputs=[inp],
        outputs=[outp],
        value_info=[mul, shape, div, matmul],
    )

    model = qonnx_make_model(graph, producer_name="mulpastconv-model")
    model = ModelWrapper(model)
    inp_values = gen_finn_dt_tensor(DataType["INT2"], [1, 4, 1, 1])
    mul_values = np.random.uniform(low=0.1, high=0.99, size=(1)).astype(np.float32)
    shape_values = np.asarray([1, -1], dtype=np.int64)
    div_values = np.random.uniform(low=0.1, high=0.99, size=(1)).astype(np.float32)
    matmul_values = gen_finn_dt_tensor(DataType["INT2"], [4, 2])
    model.set_initializer("mul", mul_values)
    model.set_initializer("shape", shape_values)
    model.set_initializer("div", div_values)
    model.set_initializer("matmul", matmul_values)
    insert_identity_op(model, op, as_first_node, approx, fork_after_id)
    model = model.transform(InferShapes())
    model = model.transform(InferShapes())
    model = model.transform(InferDataTypes())
    idict = {"inp": inp_values}
    odict_before = oxe.execute_onnx(model, idict)
    num_of_nodes_before = len(model.graph.node)
    if fork_before_id and not as_first_node:
        divout_vi = model.get_tensor_valueinfo("div_out")
        model.graph.output.append(divout_vi)
        model.graph.value_info.remove(divout_vi)
    model = model.transform(RemoveIdentityOps())
    num_of_nodes_after = len(model.graph.node)
    assert num_of_nodes_before - 1 == num_of_nodes_after

    odict_after = oxe.execute_onnx(model, idict)
    outputs_same = [np.isclose(odict_before[tname], odict_after[tname], atol=1e-3).all() for tname in odict_before.keys()]
    assert all(outputs_same)



# Copyright (c) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
import onnx
import onnx.numpy_helper as nph
import onnx.parser as oprs
from onnx.checker import check_model
from pkgutil import get_data

import qonnx.core.onnx_exec as oxe
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.general.quant import quant
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.resize_conv_to_deconv import ResizeConvolutionToDeconvolution
from qonnx.util.basic import gen_finn_dt_tensor

np.random.seed(0)


@pytest.mark.parametrize("maintain_bit_width", [True, False])
def test_resize_conv_to_deconv_float_model(maintain_bit_width: bool):
    raw_m = get_data("qonnx.data", "onnx/bsd300x3-espcn/nn_resize/float_model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    iname = model.graph.input[0].name
    oname = model.graph.output[0].name
    ishape = model.get_tensor_shape(iname)
    rand_inp = gen_finn_dt_tensor(DataType["FLOAT32"], ishape)
    input_dict = {iname: rand_inp}
    expected = oxe.execute_onnx(model, input_dict)[oname]
    new_model = model.transform(ResizeConvolutionToDeconvolution(maintain_bit_width=maintain_bit_width))
    # check that there are no Resize ops left
    op_types = list(map(lambda x: x.op_type, new_model.graph.node))
    assert "Resize" not in op_types, "Error: the Resize nodes should be removed."
    produced = oxe.execute_onnx(new_model, input_dict)[oname]
    assert np.isclose(expected, produced, atol=1e-4).all(), "Error: expected output does not match the produced output."


@pytest.mark.parametrize("maintain_bit_width", [True, False])
def test_resize_conv_to_deconv_quant_model(maintain_bit_width: bool):
    # get raw quantized model with reference input
    raw_i = get_data("qonnx.data", "onnx/bsd300x3-espcn/test_data/input_0.pb")
    raw_m = get_data("qonnx.data", "onnx/bsd300x3-espcn/nn_resize/quant_model.onnx")
    # create model from the onnx file and infer the shapes
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    iname = model.graph.input[0].name
    oname = model.graph.output[0].name
    ishape = model.get_tensor_shape(iname)
    # load the reference input tensor
    input_tensor = onnx.load_tensor_from_string(raw_i)
    input_tensor = nph.to_array(input_tensor)
    assert list(input_tensor.shape) == ishape, "Error: reference input doesn't match loaded model."
    input_dict = {iname: input_tensor}
    # get the output from the sub-pixel convolution model
    output_resize_conv = oxe.execute_onnx(model, input_dict)[oname]
    # translate the sub-pixel convolution to the deconvolution
    new_model = model.transform(ResizeConvolutionToDeconvolution(maintain_bit_width=maintain_bit_width))
    # check that there are no Resize ops left
    op_types = list(map(lambda x: x.op_type, new_model.graph.node))
    assert "Resize" not in op_types, "Error: the Resize nodes should be removed."
    # get the output from the deconvolution model
    output_deconv = oxe.execute_onnx(new_model, input_dict)[oname]
    # maintaining the specified bit width introduces additional clipping errors that
    # shouldn't be expected to maintain reasonable functional similarity
    if not maintain_bit_width:
        assert np.isclose(
            output_deconv, output_resize_conv, atol=1 / 255.0, rtol=1.0
        ).all(), "Error: expected output does not match the produced output."


def float_nn_resize_model(r: int, ifm: int, ich: int, och: int, ksize: int, use_bias: bool):
    assert isinstance(ksize, int), "Assuming square kernels, so kernel_size needs to be an int."
    pad = (ksize - 1) // 2

    ishp = (1, ich, ifm, ifm)
    oshp = (1, och, ifm * r, ifm * r)
    wshp = (och, ich, ksize, ksize)
    bshp = (och,)
    rscales = np.array([1.0, 1.0, r, r], dtype=np.float32)
    weight = np.random.randn(*wshp)
    bias = np.random.randn(*bshp)
    ishp_str = str(list(ishp))
    oshp_str = str(list(oshp))
    wshp_str = str(list(wshp))
    bshp_str = str(list(bshp))

    if use_bias:
        params_str = f"""
        <
            float{wshp_str} conv_param,
            float{bshp_str} bias_param,
            float roi,
            float scales
        >
        """
    else:
        params_str = f"""
        <
            float{wshp_str} conv_param,
            float roi,
            float scales
        >
        """

    if use_bias:
        conv_str = f"""
            out0 = Conv<
                dilations=[1,1],
                group=1,
                kernel_shape=[{ksize},{ksize}],
                strides=[1,1],
                pads=[{pad},{pad},{pad},{pad}]
            >(hid0, conv_param, bias_param)
        """
    else:
        conv_str = f"""
            out0 = Conv<
                dilations=[1,1],
                group=1,
                kernel_shape=[{ksize},{ksize}],
                strides=[1,1],
                pads=[{pad},{pad},{pad},{pad}]
            >(hid0, conv_param)
        """

    input = f"""
    <
        ir_version: 7,
        opset_import: ["" : 13]
    >
    agraph (float{ishp_str} in0) => (float{oshp_str} out0)
    {params_str}
    {{
        hid0 = Resize<
            mode="nearest"
        >(in0, roi, scales)
        {conv_str}
    }}
    """

    model = oprs.parse_model(input)
    model = ModelWrapper(model)
    model.set_initializer("roi", np.empty(0))
    model.set_initializer("scales", rscales.astype(np.float32))
    model.set_initializer("conv_param", weight.astype(np.float32))
    if use_bias:
        model.set_initializer("bias_param", bias.astype(np.float32))
    model = model.transform(InferShapes())
    check_model(model._model_proto)
    return model


def quant_nn_resize_model(r: int, ifm: int, ich: int, och: int, ksize: int, use_bias: bool, channelwise: bool):
    assert isinstance(ksize, int), "Assuming square kernels, so kernel_size needs to be an int."
    pad = (ksize - 1) // 2

    ishp = (1, ich, ifm, ifm)
    oshp = (1, och, ifm * r, ifm * r)
    wshp = (och, ich, ksize, ksize)
    bshp = (och,)
    rscales = np.array([1.0, 1.0, r, r], dtype=np.float32)
    weight = np.random.randn(*wshp)
    bias = np.random.randn(*bshp)
    ishp_str = str(list(ishp))
    oshp_str = str(list(oshp))
    wshp_str = str(list(wshp))
    bshp_str = str(list(bshp))

    if channelwise:
        q_attr_shp = (och, 1, 1, 1)
    else:
        q_attr_shp = (1,)
    attrshp_str = str(list(q_attr_shp))
    scale = np.random.rand(*q_attr_shp).astype(np.float32)
    zeropt = np.zeros(q_attr_shp).astype(np.float32)  # NOTE: needs to be integer
    bitwidth = np.array(4.0)

    weight: np.ndarray = quant(weight, scale, zeropt, bitwidth, signed=True, narrow=True, rounding_mode="ROUND")

    if use_bias:
        params_str = f"""
        <
            float{wshp_str} conv_param,
            float{attrshp_str} scale_param,
            float{attrshp_str} zeropt_param,
            float{bshp_str} bias_param,
            float bitwidth_param,
            float scale_bias,
            float zeropt_bias,
            float bitwidth_bias,
            float roi,
            float scales
        >
        """
    else:
        params_str = f"""
        <
            float{wshp_str} conv_param,
            float{attrshp_str} scale_param,
            float{attrshp_str} zeropt_param,
            float roi,
            float scales,
            float bitwidth_param
        >
        """

    if use_bias:
        scale_bias = np.random.rand(
            1,
        )
        zeropt_bias = np.array(0.0)
        bitwidth_bias = np.array(16.0)
        convs_str = f"""
        param1 = qonnx.custom_op.general.Quant<
            signed=1,
            narrow=1,
            rounding_mode="ROUND"
        >(bias_param, scale_bias, zeropt_bias, bitwidth_bias)
        out0 = Conv<
            dilations=[1,1],
            group=1,
            kernel_shape=[{ksize},{ksize}],
            strides=[1,1],
            pads=[{pad},{pad},{pad},{pad}]
        >(hid0, param0, param1)
        """
    else:
        convs_str = f"""
        out0 = Conv<
            dilations=[1,1],
            group=1,
            kernel_shape=[{ksize},{ksize}],
            strides=[1,1],
            pads=[{pad},{pad},{pad},{pad}]
        >(hid0, param0)
        """

    input = f"""
    <
        ir_version: 7,
        opset_import: ["" : 13, "qonnx.custom_op.general" : 1]
    >
    agraph (float{ishp_str} in0) => (float{oshp_str} out0)
    {params_str}
    {{
        hid0 = Resize<
            mode="nearest"
        >(in0, roi, scales)
        param0 = qonnx.custom_op.general.Quant<
            signed=1,
            narrow=1,
            rounding_mode="ROUND"
        >(conv_param, scale_param, zeropt_param, bitwidth_param)
        {convs_str}
    }}
    """
    model = oprs.parse_model(input)
    model = ModelWrapper(model)
    model.set_initializer("roi", np.empty(0))
    model.set_initializer("scales", rscales.astype(np.float32))
    model.set_initializer("conv_param", weight.astype(np.float32))
    if use_bias:
        model.set_initializer("bias_param", bias.astype(np.float32))
        model.set_initializer("scale_bias", scale_bias.astype(np.float32))
        model.set_initializer("zeropt_bias", zeropt_bias.astype(np.float32))
        model.set_initializer("bitwidth_bias", bitwidth_bias.astype(np.float32))
    model.set_initializer("scale_param", scale.astype(np.float32))
    model.set_initializer("zeropt_param", zeropt.astype(np.float32))
    model.set_initializer("bitwidth_param", bitwidth.astype(np.float32))
    model = model.transform(InferShapes())
    check_model(model._model_proto)
    return model


@pytest.mark.parametrize("kernel_size", [3, 5, 7])
@pytest.mark.parametrize("upscale_factor", [1, 2, 3, 4])
@pytest.mark.parametrize("bias", [True, False])
def test_float_resize_conv_to_deconv_layer(kernel_size: int, upscale_factor: int, bias: bool):
    och = 10  # output channels
    ich = 3  # input channels
    ifm = 4  # input feature map size
    input_shape = [1, ich, ifm, ifm]
    # Create resize convolution layer that upsamples a 4x4 image with 1 I/O channel
    model_1 = float_nn_resize_model(upscale_factor, ifm, ich, och, kernel_size, bias)
    model_2 = model_1.transform(ResizeConvolutionToDeconvolution())
    inp_dict = {"inp": np.random.rand(*input_shape).astype(np.float32)}
    assert oxe.compare_execution(model_1, model_2, inp_dict)


@pytest.mark.parametrize("kernel_size", [3, 5, 7])
@pytest.mark.parametrize("upscale_factor", [1, 2, 3, 4])
@pytest.mark.parametrize("bias", [True, False])
@pytest.mark.parametrize("channelwise", [True, False])
@pytest.mark.parametrize("maintain_bit_width", [True, False])
def test_quant_resize_conv_to_deconv_layer(
    kernel_size: int, upscale_factor: int, bias: bool, channelwise: bool, maintain_bit_width: bool
):
    och = 10  # output channels
    ich = 3  # input channels
    ifm = 4  # input feature map size
    input_shape = [1, ich, ifm, ifm]
    # Create resize convolution layer that upsamples a 4x4 image with 1 I/O channel
    model_1 = quant_nn_resize_model(upscale_factor, ifm, ich, och, kernel_size, bias, channelwise)
    model_2 = model_1.transform(ResizeConvolutionToDeconvolution(maintain_bit_width=maintain_bit_width))
    inp_dict = {"inp": np.random.rand(*input_shape).astype(np.float32)}
    assert oxe.compare_execution(model_1, model_2, inp_dict)

    if maintain_bit_width:
        bw1 = model_1.get_initializer("bitwidth_param")
        bw2 = model_2.get_initializer("bitwidth_param")
        assert (bw1 == bw2).all()



# Copyright (c) 2022 Advanced Micro Devices, Inc.
# Copyright (c) 2020-21 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of AMD nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx

import qonnx.core.data_layout as DataLayout

# optypes with the (elementwise) monotonic property
monotonic_optypes = {
    "Identity",
    "Relu",
    "LeakyRelu",
    "Clip",
    "Selu",
    "Celu",
    "Elu",
    "Sigmoid",
    "HardSigmoid",
    "Tanh",
    "Softplus",
    "Exp",
    "Log",
    "Sqrt",
    "Erf",
    "Floor",
    "Ceil",
    "Round",
    "Sign",
}

# optypes that operate in an elementwise fashion
# (with numpy-style broadcasting when shapes mismatch for binary ops)
eltwise_optypes = monotonic_optypes | {
    "Quant",
    "Mul",
    "Div",
    "Sub",
    "Add",
    "Mod",
    "And",
    "Or",
    "Xor",
    "Equal",
    "Less",
    "LessOrEqual",
    "Greater",
    "GreaterOrEqual",
    "BitwiseAnd",
    "BitwiseOr",
    "BitwiseXor",
    "Maximum",
    "Minimum",
    "BitShift",
    "Pow",
}


def is_eltwise_monotonic_optype(optype):
    "Checks whether given ONNX optype is a monotonic elementwise op."
    return optype in monotonic_optypes


def is_eltwise_optype(optype):
    "Checks whether given ONNX optype is an elementwise op."
    return optype in eltwise_optypes


def valueinfo_to_tensor(vi):
    """Creates an all-zeroes numpy tensor from a ValueInfoProto."""

    dims = [x.dim_value for x in vi.type.tensor_type.shape.dim]
    return np.zeros(dims, dtype=onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[vi.type.tensor_type.elem_type])


def nchw_to_nhwc(t, model, idx, reverse=False):
    """Converts between NCHW <-> NHWC layouts for tensor t by inserting a transpose.
    If reverse=False, t is assumed NCHW and we insert transpose to convert NCHW -> NHWC
    If reverse=True, t is assumed NHWC and we insert transpose to convert NHWC -> NCHW.
    """
    graph = model.graph
    # create new NHWC tensor
    t_shape = model.get_tensor_shape(t)
    bs = t_shape[0]
    ch = t_shape[1]
    height = t_shape[2]
    width = t_shape[3]
    t_trans = onnx.helper.make_tensor_value_info(
        model.make_new_valueinfo_name(),
        onnx.TensorProto.FLOAT,
        (bs, height, width, ch),  # NHWC
    )
    graph.value_info.append(t_trans)
    dt = model.get_tensor_datatype(t)
    t_trans = t_trans.name
    model.set_tensor_datatype(t_trans, dt)
    model.set_tensor_layout(t_trans, DataLayout.NHWC)
    # NCHW <-> NHWC transpose
    if reverse:
        t_trans_node = onnx.helper.make_node("Transpose", [t_trans], [t], perm=[0, 3, 1, 2])
    else:
        t_trans_node = onnx.helper.make_node("Transpose", [t], [t_trans], perm=[0, 2, 3, 1])
    graph.node.insert(idx, t_trans_node)
    return t_trans



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
import onnx
import os
import urllib.request as ureq

import qonnx.core.onnx_exec as oxe
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.batchnorm_to_affine import BatchNormToAffine
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import gen_finn_dt_tensor, qonnx_make_model

download_url = "https://github.com/onnx/models/raw/main/validated/vision/classification"
download_url += "/shufflenet/model/shufflenet-9.onnx"
export_onnx_path = download_url.split("/")[-1]


def test_batchnorm_to_affine_shufflenet():
    ureq.urlretrieve(download_url, export_onnx_path)
    if not os.path.isfile(export_onnx_path):
        pytest.skip("Couldn't download ONNX model, skipping")
    model = ModelWrapper(export_onnx_path)
    model = model.transform(InferShapes())
    model = model.transform(FoldConstants())
    iname = model.graph.input[0].name
    oname = model.graph.output[0].name
    ishape = model.get_tensor_shape(iname)
    np.random.seed(0)
    rand_inp = gen_finn_dt_tensor(DataType["INT8"], ishape)
    input_dict = {iname: rand_inp}
    expected = oxe.execute_onnx(model, input_dict)[oname]
    new_model = model.transform(BatchNormToAffine())
    # check that there are no BN nodes left
    op_types = list(map(lambda x: x.op_type, new_model.graph.node))
    assert "BatchNormalization" not in op_types
    produced = oxe.execute_onnx(new_model, input_dict)[oname]
    assert np.isclose(expected, produced, atol=1e-05).all()
    os.remove(export_onnx_path)


@pytest.mark.parametrize("epsilon", [0.0, 0.00001, 0.001])
def test_batchnorm_to_affine_epsilon(epsilon):
    """Dummy batchnorm node to test out the epsilon attribute."""

    batchnorm_node = onnx.helper.make_node(
        "BatchNormalization",
        inputs=["x", "s", "bias", "mean", "var"],
        outputs=["y"],
        epsilon=epsilon,
    )

    x = onnx.helper.make_tensor_value_info("x", onnx.TensorProto.FLOAT, [1, 3, 5, 5])
    s = onnx.helper.make_tensor_value_info("s", onnx.TensorProto.FLOAT, [3])
    bias = onnx.helper.make_tensor_value_info("bias", onnx.TensorProto.FLOAT, [3])
    mean = onnx.helper.make_tensor_value_info("mean", onnx.TensorProto.FLOAT, [3])
    var = onnx.helper.make_tensor_value_info("var", onnx.TensorProto.FLOAT, [3])
    y = onnx.helper.make_tensor_value_info("y", onnx.TensorProto.FLOAT, [1, 3, 5, 5])

    # Graph
    graph = onnx.helper.make_graph(
        nodes=[batchnorm_node],
        name="test_batchnorm_graph",
        inputs=[x],
        outputs=[y],
        value_info=[s, bias, mean, var],
    )

    onnx_model = qonnx_make_model(graph, producer_name="test_batchnorm-model")
    model = ModelWrapper(onnx_model)

    model.set_initializer("s", np.array([1, 2, 3]).astype(np.float32))
    model.set_initializer("bias", np.array([1, 2, 3]).astype(np.float32))
    model.set_initializer("mean", np.array([3, 4, 5]).astype(np.float32))
    model.set_initializer("var", np.array([0.5, 0.7, 0.3]).astype(np.float32))

    i_val = np.arange(0, 3 * 5 * 5, dtype=np.float32)
    i_val = np.reshape(i_val, [1, 3, 5, 5])
    input_dict = {"x": i_val}
    output_node_name = "y"

    output_dict = oxe.execute_onnx(model, input_dict, return_full_exec_context=True)
    output_original = output_dict[output_node_name]

    model_lowered = model.transform(BatchNormToAffine())
    output_dict = oxe.execute_onnx(model_lowered, input_dict, return_full_exec_context=True)
    output_lowered = output_dict[output_node_name]

    assert (output_original == output_lowered).all()



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch
from abc import ABC
from brevitas.export.common.handler.qcdq import (
    DQMixin,
    QCDQActQuantProxyHandlerMixin,
    QCDQBiasQuantProxyHandlerMixin,
    QCDQDecoupledWeightQuantProxyHandlerMixin,
    QCDQMixin,
    QCDQTruncQuantProxyHandlerMixin,
    QCDQWeightQuantProxyHandlerMixin,
)
from brevitas.export.onnx.handler import ONNXBaseHandler, QuantLSTMLayerHandler

from ..function import BrevitasQuantLSTMCellFn, DequantizeLinearFn, IntClipFn, QuantizeLinearFn


class StdDQONNXMixin(DQMixin, ABC):
    def dequantize_fn(self, x, scale, zero_point, axis):
        return DequantizeLinearFn.apply(x, scale, zero_point, axis)

    @property
    def flatten_dequantize_params(self):
        return True

    @property
    def itemize_quantize_scalar_params(self):
        return False


class StdQCDQONNXMixin(QCDQMixin, StdDQONNXMixin, ABC):
    @property
    def clip_over_integers(self):
        return True

    @classmethod
    def int8_dtype(cls):
        return torch.int8

    @classmethod
    def uint8_dtype(cls):
        return torch.uint8

    @classmethod
    def int32_dtype(cls):
        return torch.int32

    def validate(self, module):
        self.validate_8b_bit_width(module.bit_width(), le_then=True)
        assert module.bit_width() > 1.0, "Binary quant not supported"
        assert module.rounding_mode.upper() == "ROUND", "Only round to nearest even supported"

    def quantize_fn(self, x, scale, zero_point, dtype, axis):
        return QuantizeLinearFn.apply(x, scale, zero_point, dtype, axis)

    def clip_fn(self, x, min_val, max_val):
        return IntClipFn.apply(x, min_val, max_val)


class StdQCDQONNXWeightQuantProxyHandler(StdQCDQONNXMixin, QCDQWeightQuantProxyHandlerMixin, ONNXBaseHandler):
    pass


class StdQCDQONNXDecoupledWeightQuantProxyHandler(
    StdQCDQONNXMixin, QCDQDecoupledWeightQuantProxyHandlerMixin, ONNXBaseHandler
):
    pass


class StdQCDQONNXActQuantProxyHandler(StdQCDQONNXMixin, QCDQActQuantProxyHandlerMixin, ONNXBaseHandler):
    pass


class StdQCDQONNXBiasQuantProxyHandler(StdDQONNXMixin, QCDQBiasQuantProxyHandlerMixin, ONNXBaseHandler):
    pass


class StdQCDQONNXTruncQuantProxyHandler(StdQCDQONNXMixin, QCDQTruncQuantProxyHandlerMixin, ONNXBaseHandler):
    pass


class StdQCDQONNXQuantLSTMLayerHandler(QuantLSTMLayerHandler):
    def quantized_cell_symbolic_execution(
        self,
        quant_input,
        quant_hidden_state,
        quant_cell_state,
        quant_weight_ii,
        quant_weight_if,
        quant_weight_ic,
        quant_weight_io,
        quant_weight_hi,
        quant_weight_hf,
        quant_weight_hc,
        quant_weight_ho,
        quant_bias_input,
        quant_bias_forget,
        quant_bias_cell,
        quant_bias_output,
    ):
        return BrevitasQuantLSTMCellFn.apply(
            quant_input,
            quant_hidden_state,
            quant_cell_state,
            quant_weight_ii,
            quant_weight_if,
            quant_weight_ic,
            quant_weight_io,
            quant_weight_hi,
            quant_weight_hf,
            quant_weight_hc,
            quant_weight_ho,
            quant_bias_input,
            quant_bias_forget,
            quant_bias_cell,
            quant_bias_output,
            *self.symbolic_kwargs.values()
        )
        # raise RuntimeError(
        #     "Quantized LSTM cell is not supported for ONNX QCDQ "
        #     "(weights only quantization is). Use export_qonnx.")



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from qonnx.core.datatype import DataType
from qonnx.core.onnx_exec import execute_onnx
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.infer_datatypes import InferDataTypes
from qonnx.transformation.pruning import (
    ApplyMasks,
    PropagateMasks,
    PruneChannels,
    RemoveMaskedChannels,
    remove_masked_tensor_channels,
)
from qonnx.util.cleanup import cleanup_model
from qonnx.util.inference_cost import inference_cost
from qonnx.util.test import download_model, get_golden_in_and_output


def test_remove_masked_tensor_channels():
    shp = (4, 5, 6)
    x = np.random.rand(*shp)
    assert remove_masked_tensor_channels(x, [0, 2], axis=0).shape == (2, 5, 6)
    assert remove_masked_tensor_channels(x, [3], axis=1).shape == (4, 4, 6)
    assert remove_masked_tensor_channels(shp, [3], axis=1) == (4, 4, 6)


def test_apply_and_propagate_masks():
    model = download_model("FINN-TFC_W2A2", do_cleanup=True, return_modelwrapper=True)
    # manifest quantized weights as initializers
    model = model.transform(FoldConstants([]))
    mm_nodes = model.get_nodes_by_op_type("MatMul")
    # mark channels 0 and 3 in axis 1 for tensor Mul_0_out0
    # and channel 6 for the input to the 2nd MatMul as well as
    # channel 2 of input and 5 of output from the matmul weight
    # to be pruned
    mm0_input = mm_nodes[0].input[0]
    mm0_weight = mm_nodes[0].input[1]
    mm1_input = mm_nodes[1].input[0]
    mm1_weight = mm_nodes[1].input[1]
    prune_spec = {"Mul_0_out0": {1: {0, 3}}, mm0_weight: {0: {2}, 1: {5}}, mm1_input: {1: {6}}}
    model = model.transform(ApplyMasks(prune_spec))
    assert model.get_tensor_sparsity("Mul_0_out0") == prune_spec["Mul_0_out0"]
    assert model.get_tensor_sparsity(mm1_input) == prune_spec[mm1_input]
    assert model.get_tensor_sparsity(mm0_weight) == prune_spec[mm0_weight]
    # now apply the propagation
    model = model.transform(PropagateMasks())
    assert model.get_tensor_sparsity("Mul_0_out0") == {1: {0, 2, 3}}
    assert model.get_tensor_sparsity(mm0_weight) == {0: {0, 2, 3}, 1: {5, 6}}
    assert model.get_tensor_sparsity("BatchNormalization_0_out0") == {1: {5, 6}}
    model = model.transform(RemoveMaskedChannels())
    assert tuple(model.get_tensor_shape(mm0_input)) == (1, 781)
    assert tuple(model.get_tensor_shape(mm0_weight)) == (781, 62)
    assert tuple(model.get_tensor_shape(mm1_input)) == (1, 62)
    assert tuple(model.get_tensor_shape(mm1_weight)) == (62, 64)


def test_pruning_mnv1():
    model = download_model("MobileNetv1-w4a4", return_modelwrapper=True)
    # mark input as scaled 8-bit to get correct inference cost
    model.set_tensor_datatype(model.graph.input[0].name, DataType["SCALEDINT<8>"])
    model = model.transform(InferDataTypes())
    # do cleanup including folding quantized weights
    model = cleanup_model(model, False)
    inp, golden = get_golden_in_and_output("MobileNetv1-w4a4")
    cost0 = inference_cost(model, discount_sparsity=False)["total_cost"]
    assert cost0["op_mac_SCALEDINT<8>_SCALEDINT<8>"] == 10645344.0
    assert cost0["mem_w_SCALEDINT<8>"] == 864.0
    assert cost0["op_mac_SCALEDINT<4>_SCALEDINT<4>"] == 556357408.0
    assert cost0["mem_w_SCALEDINT<4>"] == 4208224.0
    prune_spec = {
        "Quant_0_out0": {1: {4, 6, 10, 13, 15, 16, 19, 26, 28}},
        "Quant_1_out0": {1: {0, 4, 6, 10, 15, 19, 26, 28}},
        "Quant_2_out0": {1: {42}},
        "Quant_3_out0": {1: {42}},
        "Quant_6_out0": {1: {102}},
        "Quant_7_out0": {1: {102}},
    }

    model = model.transform(PruneChannels(prune_spec))
    cost1 = inference_cost(model, discount_sparsity=False)["total_cost"]
    assert cost1["op_mac_SCALEDINT<8>_SCALEDINT<8>"] == 7318674.0
    assert cost1["mem_w_SCALEDINT<8>"] == 594.0
    assert cost1["op_mac_SCALEDINT<4>_SCALEDINT<4>"] == 546053216.0
    assert cost1["mem_w_SCALEDINT<4>"] == 4206942.0
    iname = model.graph.input[0].name
    oname = model.graph.output[0].name
    ret = execute_onnx(model, {iname: inp})[oname]
    assert np.isclose(golden, ret).all()



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np

import qonnx.util.basic as util
from qonnx.core.datatype import DataType
from qonnx.custom_op.base import CustomOp

# adapted from A. Karpathy's CS231 im2col code
# utilities to generate a patch matrix from a multichannel image
# of shape (batches, channels, height, width)
# note: the spatial dimensions can be set to 1 to indicate
# a dummy dimension (e.g. 1D convs represented as 2D)


def compute_conv_output_dim(ifm_dim, k, stride, total_pad=0, dilation=1):
    """Returns spatial output dimension size for convolution with given params.
    total_pad gives the total amount of padding along the entire axis
    (both sides included).
    """

    out_dim = int(((ifm_dim + total_pad - dilation * (k - 1) - 1) / stride) + 1)
    return out_dim


def get_im2col_indices_nchw(
    x_shape,
    field_height,
    field_width,
    padding=0,
    stride_h=1,
    stride_w=1,
    dilation_h=1,
    dilation_w=1,
):
    """Returns im2col indices."""
    # First figure out what the size of the output should be
    n, c, h, w = x_shape
    pad_h = padding[0] + padding[2]
    pad_w = padding[1] + padding[3]
    out_height = compute_conv_output_dim(h, field_height, stride_h, pad_h, dilation_h)
    out_width = compute_conv_output_dim(w, field_width, stride_w, pad_w, dilation_w)

    i0 = dilation_h * np.repeat(np.arange(field_height), field_width)
    i0 = np.tile(i0, c)
    i1 = stride_h * np.repeat(np.arange(out_height), out_width)
    j0 = dilation_w * np.tile(np.arange(field_width), field_height * c)
    j1 = stride_w * np.tile(np.arange(out_width), out_height)
    i = i0.reshape(-1, 1) + i1.reshape(1, -1)
    j = j0.reshape(-1, 1) + j1.reshape(1, -1)

    k = np.repeat(np.arange(c), field_height * field_width).reshape(-1, 1)

    return (k, i, j)


def im2col_indices_nchw(
    x,
    ifm_h,
    ifm_w,
    field_height,
    field_width,
    padding=[0, 0, 0, 0],
    stride_h=1,
    stride_w=1,
    pad_val=0,
    dilation_h=1,
    dilation_w=1,
):
    """Performs im2col on image (2D tensor, possibly with 1-length dummy dimensions) x with
    given field height and width, as well as values for padding and stride size.
    Returns result of im2col."""
    # Zero-pad the input
    p = padding

    x_padded = np.pad(
        x,
        ((0, 0), (0, 0), (p[0], p[2]), (p[1], p[3])),
        mode="constant",
        constant_values=pad_val,
    )

    k, i, j = get_im2col_indices_nchw(
        x.shape,
        field_height,
        field_width,
        padding,
        stride_h,
        stride_w,
        dilation_h,
        dilation_w,
    )

    cols = x_padded[:, k, i, j]
    C = x.shape[1]
    cols = cols.transpose(1, 2, 0).reshape(field_height * field_width * C, -1)
    return cols


# ONNX i/o tensor shape assumptions for Im2Col:
# input 0 is the input vector, shape (1, ih, iw, ifm)
# output 0 is the output vector, shape (1, oh, ow, kh*kw*ifm)
# where:
# * ih, iw are the height and width of the input image
# * oh, ow are the height and width of the output (lowered) image
# * ifm is the number of input channels
# * kh, kw is the convolutional kernel size

# note: for the innermost (dot product) dimension of k*k*ifm, we
# assume an internal ordering (k, k, ifm)

# note2: it's possible to set one of ih, iw to be 1 to indicate a
# dummy dimension, e.g. for representing 1D convs as 2D. the corresponding
# oh/ow and kh/kw will also be 1 in this case


class Im2Col(CustomOp):
    def get_nodeattr_types(self):
        return {
            # stride and shape of convolution kernel
            "stride": ("ints", True, []),
            "kernel_size": ("ints", True, []),
            # input tensor shape
            "input_shape": ("s", True, ""),
            # amount of padding to be inserted before/after each non-dummy spatial dim
            # i.e. [H_begin, W_begin, H_end, W_end]
            "pad_amount": ("ints", False, [0, 0, 0, 0]),  # default: no padding
            # value of padding pixels to be inserted
            "pad_value": ("i", False, 0),
            # depthwise: if 1, infer ConvolutionInputGenerator with depthwise == 1
            "depthwise": ("i", False, 0, {0, 1}),
            # dilation factor applied to the conv kernel
            "dilations": ("ints", False, [1, 1]),
        }

    def make_shape_compatible_op(self, model):
        k_h, k_w = self.get_nodeattr("kernel_size")  # Assumption: Height x Width
        stride_h, stride_w = self.get_nodeattr("stride")
        ishape = self.get_nodeattr("input_shape")
        dilation_h, dilation_w = self.get_nodeattr("dilations")
        pad = self.get_nodeattr("pad_amount")  # padding: [H_begin, W_begin, H_end, W_end]
        pad_h = pad[0] + pad[2]
        pad_w = pad[1] + pad[3]

        # convert string into list of integers
        ishape = ishape.strip("(")
        ishape = ishape.strip(")")
        ishape = ishape.split(",")
        for i in range(0, len(ishape)):
            ishape[i] = int(ishape[i])

        # extract all necessary information and determine output dimensions
        ifm_ch = ishape[-1]
        assert len(ishape) == 4, "Unexpected input shape for Im2Col"
        # NHWC (QONNX always converts to NHWC during conv lowering)
        ifm_dim_h = ishape[1]
        ifm_dim_w = ishape[2]

        ofm_dim_h = compute_conv_output_dim(ifm_dim_h, k_h, stride_h, pad_h, dilation_h)
        ofm_dim_w = compute_conv_output_dim(ifm_dim_w, k_w, stride_w, pad_w, dilation_w)

        return super().make_const_shape_op([1, ofm_dim_h, ofm_dim_w, k_h * k_w * ifm_ch])

    def infer_node_datatype(self, model):
        node = self.onnx_node
        # data type stays the same
        dtype = model.get_tensor_datatype(node.input[0])
        model.set_tensor_datatype(node.output[0], dtype)

    def execute_node(self, context, graph):
        node = self.onnx_node
        k_h, k_w = self.get_nodeattr("kernel_size")  # Assumption: Height x Width
        stride_h, stride_w = self.get_nodeattr("stride")
        pad = self.get_nodeattr("pad_amount")
        pad_h = pad[0] + pad[2]
        pad_w = pad[1] + pad[3]
        pad_val = self.get_nodeattr("pad_value")
        dilation_h, dilation_w = self.get_nodeattr("dilations")

        iname = node.input[0]
        x = context[iname]
        qnt_annotations = graph.quantization_annotation
        ret = util.get_by_name(qnt_annotations, iname, "tensor_name")
        ret = util.get_by_name(ret.quant_parameter_tensor_names, "finn_datatype", "key")
        idt = DataType[ret.value]
        if pad != [0, 0, 0, 0]:
            assert idt.allowed(pad_val), "Im2Col dtype must allow pad_val"
        # check that input is NHWC
        assert x.ndim == 4, "Unexpected number of input dims for Im2Col"
        n, h, w, c = x.shape

        # check that kernel tensor also respects any existing dummy dimensions
        if h == 1:
            kernel_1d = k_h == 1
            pad_1d = pad_h == 0
            assert (
                kernel_1d and pad_1d
            ), "Unexpected kernel shape and padding for input image\
             of dimensions (N, 1, W, C)"
        if w == 1:
            kernel_1d = k_w == 1
            pad_1d = pad_w == 0
            assert (
                kernel_1d and pad_1d
            ), "Unexpected kernel shape and padding for input image\
             of dimensions (N, H, 1, C)"

        out_dim_h = compute_conv_output_dim(h, k_h, stride_h, pad_h, dilation_h)
        out_dim_w = compute_conv_output_dim(w, k_w, stride_w, pad_w, dilation_w)
        # internally convert input to NCHW
        x = x.transpose(0, 3, 1, 2)
        # call NCHW im2col implementation
        ret = im2col_indices_nchw(
            x,
            h,
            w,
            k_h,
            k_w,
            pad,
            stride_h,
            stride_w,
            pad_val=pad_val,
            dilation_h=dilation_h,
            dilation_w=dilation_w,
        )
        # result shape is (k_H*k_W*N, out_dim_H*out_dim_W), convert to NCHW
        ret = ret.reshape(n, c, k_h, k_w, out_dim_h, out_dim_w)
        # (N=0,C=1,kh=2,kw=3,H=4,W=5) -> (N=0,H=4,W=5,kh=2,kw=3,C=1)
        ret = ret.transpose(0, 4, 5, 2, 3, 1)
        ret = ret.reshape(n, out_dim_h, out_dim_w, k_h * k_w * c)

        # ret = ret.reshape(N, k * k * C, out_dim, out_dim)
        # convert output back to NHWC
        # ret = ret.transpose(0, 2, 3, 1)
        context[node.output[0]] = ret

    def verify_node(self):
        node = self.onnx_node

        info_messages = []

        # verify number of attributes
        num_of_attr = 3
        if len(node.attribute) == num_of_attr:
            info_messages.append("The number of attributes is correct")
        else:
            info_messages.append(
                """The number of attributes is incorrect,
            {} should have {} attributes""".format(
                    node.op_type, num_of_attr
                )
            )
        # verify that all necessary attributes exist
        try:
            self.get_nodeattr("stride")
            self.get_nodeattr("kernel_size")
            info_messages.append("All necessary attributes exist")
        except Exception:
            info_messages.append(
                """The necessary attributes do not exist.
                Im2Col needs the following attributes:
                stride, kernel_size"""
            )

        # verify the number of inputs
        if len(node.input) == 1:
            info_messages.append("The number of inputs is correct")
        else:
            info_messages.append("{} needs 1 data input".format(node.op_type))

        return info_messages



import qkeras

# import tensorflow as tf
from qkeras.quantizers import BaseQuantizer
from qkeras.utils import REGISTERED_LAYERS as QKERAS_LAYERS


def extract_quantizers_from_layer(layer):
    """ """
    layer_class = layer.__class__.__name__
    if layer_class in QKERAS_LAYERS:
        handler = handler_map.get(layer_class, None)
        if handler:
            return handler_map[layer_class](layer)
        else:
            return extract_generic(layer)
    else:
        return layer_class, layer.get_config(), None


def _is_keras_quantizer(quant):
    """Check if the quantizer is a qkeras quantizer

    Args:
        quant: The quantizer node we need this information for

    Returns:
        True if the quantizer is a qkeras quantizer
    """
    try:
        # If we can deserialize the quantizer, it means it belongs to qkeras
        # TODO Since quantizer can be any callable, this should be more robust
        quant_obj = qkeras.get_quantizer(quant)
        return isinstance(quant_obj, BaseQuantizer)
    except ValueError:
        return False


def _extract_initializers(layer_cfg):
    """Return the initializers for the layer

    Args:
        layer_cfg: The layer configuration

    Returns:
        Initializers for the given layer
    """
    initializers = {}
    for key, value in layer_cfg.items():
        if value is None:
            continue
        if "initializer" in key:
            if value["class_name"] == "QInitializer":
                # Save the quantized initializer and it's replacement
                initializers[key] = (value, value["config"]["initializer"])

    return initializers


def _extract_constraints(layer_cfg):
    """Returns the constraints for the layer

    Args:
        layer_cfg: The layer configuration

    Returns:
        Initializers for the given layer
    """
    constraints = {}
    for key, value in layer_cfg.items():
        if value is None:
            continue
        if "constraint" in key:
            if value["class_name"] == "Clip":
                # QKeras doesn't keep the original constraint in the config (bug?)
                constraints[key] = (value, None)

    return constraints


_activation_map = {
    "quantized_bits": "linear",
    "quantized_relu": "relu",
    "binary": "linear",
    "ternary": "linear",
    "quantized_tanh": "tanh",
    "quantized_sigmoid": "sigmoid",
    "quantized_po2": "linear",
    "quantized_relu_po2": "relu",
}


def _replace_activation(quant_act):
    if quant_act is not None:
        for act_key, act_val in _activation_map.items():
            if act_key in quant_act:
                return act_val

    return "linear"


def extract_qlayer(layer):
    quantizers = layer.get_quantization_config()

    keras_config = layer.get_config()

    keras_config.pop("kernel_quantizer", None)
    keras_config.pop("bias_quantizer", None)
    keras_config.pop("kernel_range", None)
    keras_config.pop("bias_range", None)

    # Check if activation is quantized
    if _is_keras_quantizer(keras_config["activation"]):
        keras_config["activation"] = _replace_activation(quantizers["activation"])
    else:
        quantizers["activation"] = None

    quant_init = _extract_initializers(keras_config)
    for key, (quant, non_quant) in quant_init.items():
        quantizers[key] = quant
        keras_config[key] = non_quant

    quant_const = _extract_constraints(keras_config)
    for key, (quant, non_quant) in quant_const.items():
        quantizers[key] = quant
        keras_config[key] = non_quant

    return layer.__class__.__name__[1:], keras_config, quantizers


def extract_qact(layer):
    # As of version 0.9.0, QKeras actiations store quantizer config as a plain string, not dict
    # TODO complain to Hao about this inconsistency
    quantizers = {"activation": layer.get_quantization_config()}

    keras_config = layer.get_config()
    keras_config["activation"] = _replace_activation(quantizers["activation"])

    return "Activation", keras_config, quantizers


# This is a remnant of a first attempt at creating an universal parser
# however it got too complicated so portions were extracted to separate
# parsers. In the future, this should be removed.
def extract_generic(layer):
    get_quant_op = getattr(layer, "get_quantization_config", None)
    if callable(get_quant_op):
        quantizers = get_quant_op()
    else:
        quantizers = {}

    keras_cls_name = layer.__class__.__name__[1:]  # Drop the 'Q' from the layer name

    layer_cfg = layer.get_config()
    # Try to remove quantizers from the config
    non_quant = []
    if layer.name in quantizers:
        for quant_key, quant in quantizers.items():
            try:
                # If we can deserialize the quantizer, it means it belongs to qkeras
                qkeras.get_quantizer(quant)
                layer_cfg.pop(quant_key)
            except ValueError:
                # Otherwise it is not a quantizer (an activation, filter config, etc)
                non_quant.append(quant_key)

    for quant_key in non_quant:
        quantizers.pop(quant_key)

    # TODO Put proper activation in layer config

    # TODO extract initializers and constraints

    # Also remove deprecated 'kernel_range' and 'bias_range' From QConv1D/2D
    layer_cfg.pop("kernel_range", None)
    layer_cfg.pop("bias_range", None)

    return keras_cls_name, layer_cfg, quantizers


handler_map = {
    "QConv1D": extract_qlayer,
    "QConv2D": extract_qlayer,
    "QDense": extract_qlayer,
    "QActivation": extract_qact,
}



# Copyright (c) 2022 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx.parser as oprs

import qonnx.core.onnx_exec as oxe
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.infer_datatypes import InferDataTypes
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.lower_convs_to_matmul import LowerConvsToMatMul
from qonnx.transformation.rebalance_conv import RebalanceIm2Col
from qonnx.util.basic import gen_finn_dt_tensor


def test_rebalance_conv():
    ch_factor = 4
    ifmdim = 64
    kdim = 8
    ofmdim = ifmdim // kdim
    ifm = 1
    ofm = 16
    ishp = (1, ifm, ifmdim, ifmdim)
    oshp = (1, ofm, ofmdim, ofmdim)
    wshp = (ofm, ifm, kdim, kdim)
    dt0 = DataType["UINT8"]
    wdt = DataType["INT4"]
    np.random.seed(0)
    ishp_str = str(list(ishp))
    oshp_str = str(list(oshp))
    wshp_str = str(list(wshp))

    input = f"""
    <
        ir_version: 7,
        opset_import: ["" : 9]
    >
    agraph (float{ishp_str} in0) => (float{oshp_str} out0)
    <
        float{wshp_str} conv_param
    >
    {{
        out0 = Conv<
                dilations=[1,1], group=1, kernel_shape=[{kdim},{kdim}],
                strides=[{kdim},{kdim}], pads=[0,0,0,0]
            >(in0, conv_param)
    }}
    """
    model = oprs.parse_model(input)
    model = ModelWrapper(model)
    model.set_tensor_datatype("in0", dt0)
    w = gen_finn_dt_tensor(wdt, wshp)
    model.set_initializer("conv_param", w)
    model = model.transform(InferShapes())
    inp = gen_finn_dt_tensor(dt0, ishp)
    input_dict = {"in0": inp}
    model = model.transform(LowerConvsToMatMul())
    model = model.transform(InferShapes())
    model = model.transform(InferDataTypes())
    im2col_node = model.get_nodes_by_op_type("Im2Col")[0]
    old_im2col_ishape = model.get_tensor_shape(im2col_node.input[0])
    old_im2col_oshape = model.get_tensor_shape(im2col_node.output[0])
    assert tuple(old_im2col_ishape) == (1, ifmdim, ifmdim, ifm)
    out_expected = oxe.execute_onnx(model, input_dict)["out0"]
    model = model.transform(RebalanceIm2Col(ch_factor))
    model = model.transform(InferShapes())
    model = model.transform(InferDataTypes())
    im2col_node = model.get_nodes_by_op_type("Im2Col")[0]
    new_im2col_ishape = model.get_tensor_shape(im2col_node.input[0])
    new_im2col_oshape = model.get_tensor_shape(im2col_node.output[0])
    out_produced = oxe.execute_onnx(model, input_dict)["out0"]
    assert len(model.get_nodes_by_op_type("Reshape")) == 1
    assert tuple(new_im2col_ishape) == (1, ifmdim, ifmdim // ch_factor, ch_factor)
    assert old_im2col_oshape == new_im2col_oshape
    assert (out_expected == out_produced).all()



# Copyright (c) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import warnings
from onnx import helper

from qonnx.transformation.base import Transformation
from qonnx.util.basic import auto_pad_to_explicit_padding, get_by_name


def _weight_shuffle(cnv_weights: np.ndarray, block_size: int) -> np.ndarray:
    """Adaptation of the weight shuffle algorithm as proposed in Colbert et al. (2021) - `An
    Energy-Efficient Edge Computing Paradigm for Convolution-Based Image Upsampling`"""
    ofm_ch = cnv_weights.shape[0]
    ifm_ch = cnv_weights.shape[1]
    assert ofm_ch % block_size == 0, "Out channels need to be evenly divisible by block size"
    ofm_ch = ofm_ch // pow(block_size, 2)
    kh_size = cnv_weights.shape[2]
    kw_size = cnv_weights.shape[3]
    assert kh_size == kw_size, "Only square channels supported currently."
    # NOTE - this is different than the convolution kernels, which are OC x IC x KH x KW
    #        rather than IC x OC x KH x KW
    dcnv_weights = np.zeros((ifm_ch, ofm_ch, kh_size * block_size, kw_size * block_size))
    cnv_weights = np.moveaxis(cnv_weights, 0, 1)  # change conv to match deconv data format
    for oc_d in range(ofm_ch):
        for kh_d in range(kh_size * block_size):
            for kw_d in range(kw_size * block_size):
                kh_c = kh_d // block_size
                kw_c = kw_d // block_size
                _a = kh_d % block_size
                _b = kw_d % block_size
                _c = oc_d
                oc_c = (pow(block_size, 2) * _c) + (block_size * _a) + _b
                # shuffling and transposing the kernels
                dcnv_weights[:, oc_d, kh_d, kw_d] = cnv_weights[:, oc_c, kh_size - kh_c - 1, kw_size - kw_c - 1]
    return dcnv_weights


class SubPixelToDeconvolution(Transformation):
    """Replaces sub-pixel convolution layers (i.e., same-padded convolution + depth2space)
    with deconvolution layers using the weight shuffle algorithm. Currently does not support
    same-padded convolutions with biases."""

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        graph_modified = False
        for n in graph.node:
            node_ind += 1
            if n.op_type == "Conv":
                cnv_input = n.input[0]
                cnv_output = n.output[0]
                consumers = model.find_consumers(cnv_output)
                if len(consumers) > 1 and any([c.op_type == "DepthToSpace" for c in consumers]):
                    warnings.warn(
                        "Skipping sub-pixel conv that has same-padded conv with multiple consumers. Not yet supported."
                    )
                    continue
                consumer = consumers[0]
                # NOTE - currently supports onnx opset 11+, which introduced the
                # pixel shuffle operator as a depth2space op_type
                if consumer is not None and consumer.op_type == "DepthToSpace":
                    # TODO - converting sub-pixel convolution with bias requires a non-trivial
                    # elementwise addition node since the bias needs to be shuffled also
                    if len(n.input) == 3:
                        warnings.warn("Skipping sub-pixel conv with bias. Not yet supported.")
                        continue
                    group = get_by_name(n.attribute, "group").i
                    if group != 1:
                        warnings.warn("Skipping sub-pixel conv with group > 1. Not yet supported.")
                        continue

                    # The weights of the convolution can be generated by another input op if the model is
                    # quantized. Preliminary support for quantization focuses on QONNX ops (i.e., Quant)
                    weight_name = n.input[1]
                    weight_prod = model.find_producer(weight_name)

                    # If the producer is None, then it is initialized by the Conv node
                    if weight_prod is None:
                        W_conv = model.get_initializer(weight_name)  # (OC, IC, KH, KW)

                    # If the convolution weights are not initialized by the convolution, then we need to
                    # find the node is producing the weights
                    else:
                        if weight_prod.op_type == "Quant":
                            [q_w_name, q_s_name, _, _] = weight_prod.input
                            W_conv = model.get_initializer(q_w_name)
                            W_scale = model.get_initializer(q_s_name)
                            assert W_scale.ndim == 0, "Only supporting per-tensor quantization with this transformation."
                        else:
                            warnings.warn(
                                f"Weight producer is {weight_prod.op_type}, not a QONNX Quant node. Not yet supported."
                            )
                            continue

                    kshape = get_by_name(n.attribute, "kernel_shape").ints
                    idim = model.get_tensor_shape(n.input[0])  # require NCHW
                    odim = model.get_tensor_shape(n.output[0])  # require NCHW
                    if not (len(odim) == len(idim) == 4):
                        warnings.warn("Skipping sub-pixel conv, only 2D convolutions supported.")
                        continue

                    [_, ifm_ch, ifm_dim_h, ifm_dim_w] = idim
                    [_, ofm_ch, ofm_dim_h, ofm_dim_w] = odim

                    if (ifm_dim_h != ofm_dim_h) or (ifm_dim_w != ofm_dim_w):
                        warnings.warn("Skipping sub-pixel conv, only same-padded convs supported.")
                        continue
                    dilation_attr = get_by_name(n.attribute, "dilations")
                    if dilation_attr is not None:
                        dilation = dilation_attr.ints
                    else:
                        dilation = [1, 1]  # default value
                    if dilation != [1, 1]:
                        warnings.warn("Skipping sub-pixel conv, only supporting dilation=[1,1].")
                    # get depth-to-space (i.e., pixel shuffle) op attributes
                    block_size = get_by_name(consumer.attribute, "blocksize").i
                    if ofm_ch % block_size != 0:
                        warnings.warn(
                            "Skipping sub-pixel conv, the output channels and block size need to be evenly divisible."
                        )
                        continue
                    W_deconv = _weight_shuffle(W_conv, block_size).astype(np.float32)
                    kh_size_deconv = kshape[0] * block_size
                    kw_size_deconv = kshape[1] * block_size
                    ofm_ch_deconv = ofm_ch // pow(block_size, 2)
                    assert W_deconv.shape == (
                        ifm_ch,
                        ofm_ch_deconv,
                        kh_size_deconv,
                        kw_size_deconv,
                    ), "The resulting deconvolution weight shape is incorrect."
                    stride_h = get_by_name(n.attribute, "strides").ints[0]
                    stride_w = get_by_name(n.attribute, "strides").ints[1]
                    # handle both auto_pad and explicit padding
                    auto_pad = get_by_name(n.attribute, "auto_pad")
                    if auto_pad is not None:
                        # find equivalent specified padding
                        auto_pad = auto_pad.s.decode("utf-8")
                        if auto_pad == "NOTSET":
                            # use specified padding
                            pad = get_by_name(n.attribute, "pads").ints
                        else:
                            pad = auto_pad_to_explicit_padding(
                                auto_pad,
                                ifm_dim_h,
                                ifm_dim_w,
                                kshape[0],
                                kshape[1],
                                stride_h,
                                stride_w,
                                len(model.get_tensor_shape(n.input[0])) - 2,
                            )
                    else:
                        # use specified padding
                        pad = get_by_name(n.attribute, "pads").ints

                    # If len(pad) == 2, assume no padding for other dimension
                    if len(pad) == 2:  # only one dimension should be padded
                        assert ifm_dim_h == 1 or ifm_dim_w == 1, "Padding is assumed to be 1D, image is 2D"
                    deconv_inp = cnv_input
                    deconv_out = consumer.output[0]
                    deconv_pad = [p * block_size for p in pad]
                    deconv_node = helper.make_node(
                        "ConvTranspose",
                        [deconv_inp, weight_name],
                        [deconv_out],
                        kernel_shape=[kh_size_deconv, kw_size_deconv],
                        strides=[block_size, block_size],
                        pads=deconv_pad,
                        group=group,
                        dilations=dilation,
                    )
                    W_deconv_init = weight_name
                    if weight_prod is not None:
                        W_deconv_init = q_w_name
                    model.set_initializer(W_deconv_init, W_deconv)
                    model.set_tensor_shape(weight_name, list(W_deconv.shape))
                    graph.node.insert(node_ind, deconv_node)
                    # remove old nodes
                    graph.node.remove(n)
                    graph.node.remove(consumer)
                    graph_modified = True
        return (model, graph_modified)



# Copyright (c) 2021 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnxruntime as rt
from onnx import TensorProto, helper

from qonnx.core.datatype import DataType
from qonnx.custom_op.base import CustomOp
from qonnx.custom_op.general.maxpoolnhwc import compute_pool_output_dim
from qonnx.util.basic import qonnx_make_model


class QuantAvgPool2d(CustomOp):
    """CustomOp that corresponds to the quantized average pooling
    layer from Brevitas"""

    def get_nodeattr_types(self):
        return {
            "stride": ("i", True, 1),
            "kernel": ("i", True, 1),
            "ibits": ("i", True, 1),
            "obits": ("i", True, 1),
            # determines if values are signed (set to "1") or unsigned ("0")
            "signed": ("i", True, 0, {0, 1}),
            # data layout attribute can be set to "NCHW" or "NHWC"
            "data_layout": ("s", False, "NCHW", {"NCHW", "NHWC"}),
        }

    def make_shape_compatible_op(self, model):
        node = self.onnx_node
        k = self.get_nodeattr("kernel")
        s = self.get_nodeattr("stride")
        data_layout = self.get_nodeattr("data_layout")
        if data_layout == "NCHW":
            return helper.make_node(
                "AveragePool",
                inputs=[node.input[0]],
                outputs=[node.output[0]],
                kernel_shape=[k, k],
                strides=[s, s],
            )
        elif data_layout == "NHWC":
            iname = node.input[0]
            ishape = model.get_tensor_shape(iname)
            (n, hi, wi, c) = ishape
            ho = compute_pool_output_dim(hi, k, s)
            wo = compute_pool_output_dim(wi, k, s)
            oshape = (n, ho, wo, c)
            return super().make_const_shape_op(oshape)

        else:
            raise Exception(
                """Datalayout for QuantAvgPool2d is set to an invalid value.
                    Has to be set to "NCHW" or "NHWC"."""
            )

    def infer_node_datatype(self, model):
        node = self.onnx_node
        bw = self.get_nodeattr("obits")
        if bw in range(2, 33):
            if self.get_nodeattr("signed") == 0:
                dtype = DataType["UINT%d" % bw]
            else:
                dtype = DataType["INT%d" % bw]
        else:
            raise Exception("Unsupported output datatype for QuantAvgPool2d")
        model.set_tensor_datatype(node.output[0], dtype)

    def get_accum_size(self):
        ibits = self.get_nodeattr("ibits")
        k = self.get_nodeattr("kernel")
        max_value = 2**ibits - 1
        max_value = max_value * k * k
        max_bit_width = int(max_value).bit_length()
        return max_bit_width

    def get_shifts(self):
        shift_bits = self.get_accum_size() - self.get_nodeattr("obits")
        shift_bits = shift_bits if shift_bits >= 0 else 0
        return shift_bits

    def execute_node(self, context, graph):
        # create a standard average pooling node to help calculate the result
        node = self.onnx_node
        k = self.get_nodeattr("kernel")
        s = self.get_nodeattr("stride")
        inp_values = context[node.input[0]]
        oshape = context[node.output[0]].shape
        if self.get_nodeattr("data_layout") == "NHWC":
            inp_values = inp_values.transpose(0, 3, 1, 2)
            oshape = (context[node.output[0]]).transpose(0, 3, 1, 2).shape
        ishape = inp_values.shape
        inp = helper.make_tensor_value_info(node.input[0], TensorProto.FLOAT, ishape)
        outp = helper.make_tensor_value_info(node.output[0], TensorProto.FLOAT, oshape)
        node_avgpool = helper.make_node(
            "AveragePool",
            inputs=[node.input[0]],
            outputs=[node.output[0]],
            kernel_shape=[k, k],
            strides=[s, s],
        )
        graph_avgpool = helper.make_graph(
            nodes=[node_avgpool],
            name="single-avgpool-exec",
            inputs=[inp],
            outputs=[outp],
        )

        opset_version = self.onnx_opset_version
        opset_imports = [helper.make_opsetid("", opset_version)]
        onnx_kwargs = {"opset_imports": opset_imports}
        model_avgpool = qonnx_make_model(graph_avgpool, **onnx_kwargs)
        idict = {node.input[0]: inp_values}
        sess = rt.InferenceSession(model_avgpool.SerializeToString())
        result_temp = sess.run(None, idict)
        # remove scaling introduced by average
        result_temp = np.round(result_temp[0] * (k * k))
        result = np.right_shift(result_temp.astype(int), self.get_shifts())
        if self.get_nodeattr("data_layout") == "NHWC":
            result = result.transpose(0, 2, 3, 1)
        context[node.output[0]] = result.astype(np.float32)

    def verify_node(self):
        info_messages = []
        return info_messages



import pytest

import numpy as np
import onnx
import os
import tensorflow as tf
from qkeras import QActivation, QConv2D, QDense, binary, quantized_bits, quantized_relu, ternary
from tensorflow.keras.layers import Activation, Conv2D, Dense, Flatten, Input
from tensorflow.keras.models import Model

import qonnx.core.onnx_exec as oxe
from qonnx.converters import from_keras
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.infer_shapes import InferShapes

act_quantizers = [
    quantized_bits(8, 4, 0, alpha=1),
    quantized_bits(8, 4, 1, alpha=1),
    quantized_bits(8, 8, 0, alpha=1),
    quantized_bits(8, 8, 1, alpha=1),
    quantized_bits(4, 4, 0, alpha=1),
    quantized_bits(4, 4, 1, alpha=1),
    quantized_bits(4, 0, 0, alpha=1),
    quantized_bits(4, 0, 1, alpha=1),
    quantized_bits(4, 2, 0, alpha=1),
    quantized_bits(2, 2, 1, alpha=1),
    quantized_bits(2, 1, 1, alpha=1),
    #   ternary(alpha=1, threshold=0.5), Not stable
    binary(alpha=1),
]
act_quantizers_ids = list(range(len(act_quantizers)))

act_quantizers_relu = [
    quantized_relu(8),
    quantized_relu(8, 4),
    quantized_relu(4),
    quantized_relu(4, 4),
    quantized_relu(6, 2),
]
act_quantizers_relu_ids = list(range(len(act_quantizers_relu)))


@pytest.mark.parametrize("quantizer", act_quantizers_relu, ids=act_quantizers_relu_ids)
def test_qkeras_qactivation(quantizer, request):
    x = x_in = Input((16), name="input")
    x = QActivation(activation=quantizer, name="act_0")(x)
    model = Model(inputs=[x_in], outputs=[x])
    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 16)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_qkeras_conversion", opset=9)
    assert external_storage is None
    model_path = f"model_test_qkeras_qactivation_{request.node.callspec.id}.onnx"
    onnx.save(onnx_model, model_path)

    onnx_model = ModelWrapper(model_path)
    onnx_model = onnx_model.transform(InferShapes())

    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]

    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-5, atol=1e-5)
    os.remove(model_path)


# pairs of quantizers for kernel and bias
kb_quantizers = [
    (quantized_bits(8, 4, 0, alpha=1), quantized_bits(8, 4, 0, alpha=1)),
    (quantized_bits(8, 4, 1, alpha=1), quantized_bits(8, 4, 1, alpha=1)),
    (quantized_bits(8, 8, 0, alpha=1), quantized_bits(8, 8, 0, alpha=1)),
    (quantized_bits(8, 8, 1, alpha=1), quantized_bits(8, 8, 1, alpha=1)),
    (quantized_bits(4, 4, 0, alpha=1), quantized_bits(8, 8, 0, alpha=1)),
    (quantized_bits(4, 4, 1, alpha=1), quantized_bits(8, 8, 1, alpha=1)),
    (quantized_bits(4, 0, 0, alpha=1), quantized_bits(8, 0, 0, alpha=1)),
    (quantized_bits(4, 0, 1, alpha=1), quantized_bits(8, 0, 1, alpha=1)),
    (quantized_bits(4, 2, 0, alpha=1), quantized_bits(8, 2, 0, alpha=1)),
    (quantized_bits(2, 2, 1, alpha=1), quantized_bits(2, 2, 1, alpha=1)),
    (quantized_bits(2, 1, 1, alpha=1), quantized_bits(2, 1, 1, alpha=1)),
    (ternary(alpha=1, threshold=0.5), quantized_bits(4, 4)),
    (binary(alpha=1), quantized_bits(4, 4)),
]
kb_quantizers_ids = list(range(len(kb_quantizers)))


def test_keras_conv2d_conversion():
    x = x_in = Input((28, 28, 1), name="input")
    x = Conv2D(32, (2, 2), strides=(2, 2), name="conv2d_0_m")(x)
    x = Activation("relu", name="act0_m")(x)
    x = Conv2D(64, (3, 3), strides=(2, 2), name="conv2d_1_m", activation="relu")(x)
    x = Conv2D(64, (2, 2), strides=(2, 2), name="conv2d_2_m")(x)
    x = Activation("relu", name="act2_m")(x)
    x = Flatten(name="flatten")(x)
    x = Dense(10, bias_initializer="ones", name="dense")(x)
    x = Activation("softmax", name="softmax")(x)

    model = Model(inputs=[x_in], outputs=[x])

    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 28, 28, 1)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_keras_conv2d_conversion")
    onnx_model = onnx.shape_inference.infer_shapes(onnx_model)

    assert external_storage is None
    onnx.save(onnx_model, "model_test_keras_conv2d_conversion.onnx")

    onnx_model = ModelWrapper("model_test_keras_conv2d_conversion.onnx")
    onnx_model = onnx_model.transform(InferShapes())

    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]

    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-5, atol=1e-5)
    os.remove("model_test_keras_conv2d_conversion.onnx")


def test_keras_dense_conversion():
    ini = tf.keras.initializers.RandomUniform(minval=-1.0, maxval=1.0)
    x = x_in = Input((15), name="input")
    x = Dense(10, kernel_initializer=ini, bias_initializer=ini, name="dense1")(x)
    x = Activation("relu", name="act0_m")(x)
    x = Dense(10, kernel_initializer=ini, bias_initializer=ini, activation="relu", name="dense2")(x)
    x = Dense(10, kernel_initializer=ini, bias_initializer=ini, name="dense3")(x)
    x = Activation("softmax", name="softmax")(x)

    model = Model(inputs=[x_in], outputs=[x])

    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 15)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_keras_dense_conversion")
    assert external_storage is None
    onnx.save(onnx_model, "model_test_keras_dense_conversion.onnx")

    onnx_model = ModelWrapper("model_test_keras_dense_conversion.onnx")
    onnx_model = onnx_model.transform(InferShapes())

    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]

    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-5, atol=1e-5)
    os.remove("model_test_keras_dense_conversion.onnx")


@pytest.mark.parametrize("quantizers", kb_quantizers, ids=kb_quantizers_ids)
def test_qkeras_qdense_1(quantizers, request):
    kq, bq = quantizers
    # Initialize the kernel & bias to RandonUniform within the range of the quantizers
    k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
    b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
    x = x_in = Input((16), name="input")
    x = QDense(
        32,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense_0",
    )(x)
    model = Model(inputs=[x_in], outputs=[x])
    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 16)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_qkeras_conversion", opset=9)
    assert external_storage is None
    model_path = f"model_test_qkeras_qdense1_{request.node.callspec.id}.onnx"
    onnx.save(onnx_model, model_path)

    onnx_model = ModelWrapper(model_path)
    onnx_model = onnx_model.transform(InferShapes())

    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]

    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)
    os.remove(model_path)


@pytest.mark.parametrize("quantizers", kb_quantizers, ids=kb_quantizers_ids)
def test_qkeras_qdense_2(quantizers, request):
    kq, bq = quantizers
    # Initialize the kernel & bias to RandonUniform within the range of the quantizers
    k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
    b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
    x = x_in = Input((16), name="input")
    x = QDense(
        32,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense_0",
    )(x)
    x = QActivation("quantized_relu(6,2)", name="act1")(x)
    x = QDense(
        32,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense_1",
    )(x)
    model = Model(inputs=[x_in], outputs=[x])
    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 16)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_qkeras_conversion", opset=9)
    assert external_storage is None
    model_path = f"model_test_qkeras_qdense2_{request.node.callspec.id}.onnx"
    onnx.save(onnx_model, model_path)

    onnx_model = ModelWrapper(model_path)
    onnx_model = onnx_model.transform(InferShapes())

    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]
    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)
    os.remove(model_path)


@pytest.mark.parametrize("quantizers", kb_quantizers, ids=kb_quantizers_ids)
def test_qkeras_qdense_3(quantizers, request):
    kq, bq = quantizers
    # Initialize the kernel & bias to RandonUniform within the range of the quantizers
    k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
    b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
    x = x_in = Input((16), name="input")
    x = QDense(
        32,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        activation=quantized_bits(6, 2, 1, alpha=1.0),
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense_0",
    )(x)
    x = QActivation("quantized_relu(6,4,0)", name="act2_m")(x)
    x = QDense(
        10,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense_1",
    )(x)
    x = Activation("softmax", name="softmax")(x)
    model = Model(inputs=[x_in], outputs=[x])
    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 16)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_qkeras_conversion", opset=9)
    assert external_storage is None
    model_path = f"model_test_qkeras_qdense3_{request.node.callspec.id}.onnx"
    onnx.save(onnx_model, model_path)

    onnx_model = ModelWrapper(model_path)
    onnx_model = onnx_model.transform(InferShapes())

    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]
    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)
    os.remove(model_path)


@pytest.mark.parametrize("quantizers", act_quantizers_relu, ids=act_quantizers_relu_ids)
def test_qkeras_qdense_4(quantizers, request):
    kq, bq = (quantized_bits(4, 0, 1, alpha=1), quantized_bits(8, 0, 1, alpha=1))
    # Initialize the kernel & bias to RandonUniform within the range of the quantizers
    k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
    b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
    x = x_in = Input((16), name="input")
    x = QDense(
        32,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        activation=quantized_bits(8, 8, 1, alpha=1.0),
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense_0",
    )(x)
    x = QActivation(activation=quantizers, name="act_1")(x)
    x = QDense(
        32,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense_1",
    )(x)
    x = QActivation(activation=quantizers, name="act_2")(x)
    x = QDense(
        10,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        activation=quantized_bits(4, 4, 0, alpha=1.0),
        use_bias=False,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense_2",
    )(x)
    x = Activation("softmax", name="softmax")(x)
    model = Model(inputs=[x_in], outputs=[x])
    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 16)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_qkeras_conversion", opset=9)
    assert external_storage is None
    model_path = f"model_test_qkeras_qdense4_{request.node.callspec.id}.onnx"
    onnx.save(onnx_model, model_path)

    onnx_model = ModelWrapper(model_path)
    onnx_model = onnx_model.transform(InferShapes())

    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]
    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)
    os.remove(model_path)


@pytest.mark.parametrize("quantizers", kb_quantizers, ids=kb_quantizers_ids)
def test_qkeras_qconv2d_1(quantizers, request):
    kq, bq = quantizers
    k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
    b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
    x = x_in = Input((28, 28, 3), name="input")
    x = QConv2D(
        32,
        (2, 2),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        activation=quantized_bits(4, 4, 1, alpha=1.0),
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_0",
    )(x)
    x = QActivation("quantized_relu(6,2)", name="act1")(x)
    x = QConv2D(
        64,
        (3, 3),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        use_bias=False,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_1",
    )(x)
    model = Model(inputs=[x_in], outputs=[x])

    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 28, 28, 3)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_qkeras_conversion", opset=9)
    assert external_storage is None
    model_path = f"model_test_qkeras_qconv2d1_{request.node.callspec.id}.onnx"
    onnx.save(onnx_model, model_path)

    onnx_model = ModelWrapper(model_path)
    onnx_model = onnx_model.transform(InferShapes())

    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]

    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)
    os.remove(model_path)


@pytest.mark.parametrize("quantizers", act_quantizers_relu, ids=act_quantizers_relu_ids)
def test_qkeras_qconv2d_2(quantizers, request):
    kq, bq = (quantized_bits(4, 0, 1, alpha=1), quantized_bits(8, 0, 1, alpha=1))
    k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
    b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
    x = x_in = Input((28, 28, 3), name="input")
    x = QConv2D(
        32,
        (2, 2),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        activation=quantized_bits(4, 4, 1, alpha=1.0),
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_0",
    )(x)
    x = QActivation(quantizers, name="act1")(x)
    x = QConv2D(
        64,
        (3, 3),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        use_bias=False,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_1",
    )(x)
    model = Model(inputs=[x_in], outputs=[x])

    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 28, 28, 3)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_qkeras_conversion", opset=9)
    assert external_storage is None
    model_path = f"model_test_qkeras_qconv2d2_{request.node.callspec.id}.onnx"
    onnx.save(onnx_model, model_path)

    onnx_model = ModelWrapper(model_path)
    onnx_model = onnx_model.transform(InferShapes())

    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]

    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)
    os.remove(model_path)


@pytest.mark.parametrize("quantizers", act_quantizers, ids=act_quantizers_ids)
def test_qkeras_qconv2d_3(quantizers, request):
    kq, bq = (quantized_bits(4, 2, 0, alpha=1), quantized_bits(8, 8, 1, alpha=1))
    k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
    b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
    x = x_in = Input((28, 28, 3), name="input")
    x = QConv2D(
        32,
        (2, 2),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        activation=quantizers,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_0",
    )(x)
    x = QConv2D(
        64,
        (3, 3),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        activation=quantizers,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_1",
    )(x)
    x = QActivation("quantized_relu(6,2)", name="act1")(x)
    x = QConv2D(
        64,
        (3, 3),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_2",
    )(x)
    model = Model(inputs=[x_in], outputs=[x])

    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 28, 28, 3)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_qkeras_conversion", opset=9)
    assert external_storage is None
    model_path = f"model_test_qkeras_qconv2d3_{request.node.callspec.id}.onnx"
    onnx.save(onnx_model, model_path)

    onnx_model = ModelWrapper(model_path)
    onnx_model = onnx_model.transform(InferShapes())

    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]

    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)
    os.remove(model_path)


@pytest.mark.parametrize("quantizers", kb_quantizers, ids=kb_quantizers_ids)
def test_qkeras_qconv2d_conversion_1(quantizers, request):
    kq, bq = quantizers
    k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
    b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
    x = x_in = Input((28, 28, 1), name="input")
    x = QConv2D(
        32,
        (2, 2),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_0_m",
    )(x)
    x = QConv2D(
        64,
        (3, 3),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_1_m",
    )(x)
    x = QConv2D(
        64,
        (2, 2),
        strides=(2, 2),
        kernel_quantizer=kq,
        kernel_initializer=k_ini,
        use_bias=False,  # Lets try this one without bias to see if that trips up the converter
        name="conv2d_2_m",
    )(x)
    x = QConv2D(
        64,
        (2, 2),
        strides=(2, 2),
        kernel_quantizer=kq,
        kernel_initializer=k_ini,
        use_bias=False,  # Lets try this one without bias to see if that trips up the converter
        name="conv2d_3_m",
    )(x)
    x = QActivation("quantized_relu(6,4)", name="act2_m")(x)
    x = Flatten(name="flatten")(x)
    x = QDense(
        10,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense",
    )(x)
    model = Model(inputs=[x_in], outputs=[x])

    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 28, 28, 1)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_qkeras_qconv2d_conversion", opset=9)
    assert external_storage is None
    model_path = f"model_test_qkeras_qconv2d_conversion1_{request.node.callspec.id}.onnx"
    onnx.save(onnx_model, model_path)

    onnx_model = ModelWrapper(model_path)
    onnx_model = onnx_model.transform(InferShapes())
    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]
    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)
    os.remove(model_path)


@pytest.mark.parametrize("quantizers", act_quantizers_relu, ids=act_quantizers_relu_ids)
def test_qkeras_qconv2d_conversion_2(quantizers, request):
    kq, bq = (quantized_bits(4, 4, 0, alpha=1), quantized_bits(8, 8, 0, alpha=1))
    k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
    b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
    x = x_in = Input((28, 28, 1), name="input")
    x = QConv2D(
        32,
        (2, 2),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        activation=quantized_bits(4, 2, 0, alpha=1),
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_0_m",
    )(x)
    x = QActivation(quantizers, name="act1_m")(x)
    x = QConv2D(
        64,
        (3, 3),
        strides=(2, 2),
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="conv2d_1_m",
    )(x)
    x = Flatten(name="flatten")(x)
    x = QActivation(quantizers, name="act2_m")(x)
    x = QDense(
        32,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        activation=quantized_bits(8, 2, 1, alpha=1),
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense_1",
    )(x)
    x = QDense(
        10,
        kernel_quantizer=kq,
        bias_quantizer=bq,
        kernel_initializer=k_ini,
        bias_initializer=b_ini,
        name="dense_2",
    )(x)
    x = Activation("softmax", name="softmax")(x)

    model = Model(inputs=[x_in], outputs=[x])

    x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 28, 28, 1)).astype(dtype=np.float32)
    y_qkeras = model.predict(x_test)

    onnx_model, external_storage = from_keras(model, "test_qkeras_qconv2d_conversion", opset=9)
    assert external_storage is None
    model_path = f"model_test_qkeras_qconv2d_conversion2_{request.node.callspec.id}.onnx"
    onnx.save(onnx_model, model_path)

    onnx_model = ModelWrapper(model_path)
    onnx_model = onnx_model.transform(InferShapes())
    idict = {onnx_model.graph.input[0].name: x_test}
    odict = oxe.execute_onnx(onnx_model, idict, True)
    y_qonnx = odict[onnx_model.graph.output[0].name]
    np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)
    os.remove(model_path)


# quantized_relu should not be used as a layer activation
# def test_qkeras_broken_1(quantizers, request):
#     kq, bq = (quantized_bits(4, 4, 0, alpha=1), quantized_bits(8, 8, 0, alpha=1))
#     # Initialize the kernel & bias to RandonUniform within the range of the quantizers
#     k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
#     b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
#     x = x_in = Input((16), name="input")
#     x = QDense(
#         32,
#         kernel_quantizer=kq,
#         bias_quantizer=bq,
#         activation="quantized_relu(6,2)",
#         kernel_initializer=k_ini,
#         bias_initializer=b_ini,
#         name="dense_0",
#     )(x)
#     x = QDense(
#         32,
#         kernel_quantizer=kq,
#         bias_quantizer=bq,
#         kernel_initializer=k_ini,
#         bias_initializer=b_ini,
#         name="dense_1",
#     )(x)
#     model = Model(inputs=[x_in], outputs=[x])
#     x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 16)).astype(dtype=np.float32)
#     y_qkeras = model.predict(x_test)

#     onnx_model, external_storage =from_keras(model, "test_qkeras_conversion", opset=9)
#     assert external_storage is None
#     model_path = f"test_qkeras_broken1{request.node.callspec.id}.onnx"
#     onnx.save(onnx_model, model_path)

#     onnx_model = ModelWrapper(model_path)
#     onnx_model = onnx_model.transform(InferShapes())

#     idict = {onnx_model.graph.input[0].name: x_test}
#     odict = oxe.execute_onnx(onnx_model, idict, True)
#     y_qonnx = odict[onnx_model.graph.output[0].name]
#     np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)


# quantized_bits should not be used in QActivation
# def test_qkeras_broken_2(quantizers, request):
#     kq, bq = (quantized_bits(4, 4, 0, alpha=1), quantized_bits(8, 8, 0, alpha=1))
#     # Initialize the kernel & bias to RandonUniform within the range of the quantizers
#     k_ini = tf.keras.initializers.RandomUniform(minval=kq.min(), maxval=kq.max())
#     b_ini = tf.keras.initializers.RandomUniform(minval=bq.min(), maxval=bq.max())
#     x = x_in = Input((16), name="input")
#     x = QDense(
#         32,
#         kernel_quantizer=kq,
#         bias_quantizer=bq,
#         kernel_initializer=k_ini,
#         bias_initializer=b_ini,
#         name="dense_0",
#     )(x)
#     x = QActivation("quantized_bits(4,4,1)", name="act2_m")(x)
#     x = QDense(
#         32,
#         kernel_quantizer=kq,
#         bias_quantizer=bq,
#         kernel_initializer=k_ini,
#         bias_initializer=b_ini,
#         name="dense_1",
#     )(x)
#     model = Model(inputs=[x_in], outputs=[x])
#     x_test = np.random.uniform(low=-1.0, high=1.0, size=(1, 16)).astype(dtype=np.float32)
#     y_qkeras = model.predict(x_test)

#     onnx_model, external_storage =from_keras(model, "test_qkeras_conversion", opset=9)
#     assert external_storage is None
#     model_path = f"test_qkeras_broken2{request.node.callspec.id}.onnx"
#     onnx.save(onnx_model, model_path)

#     onnx_model = ModelWrapper(model_path)
#     onnx_model = onnx_model.transform(InferShapes())

#     idict = {onnx_model.graph.input[0].name: x_test}
#     odict = oxe.execute_onnx(onnx_model, idict, True)
#     y_qonnx = odict[onnx_model.graph.output[0].name]
#     np.testing.assert_allclose(y_qkeras, y_qonnx, rtol=1e-4, atol=1e-4)

if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

from qonnx.transformation.expose_intermediate import ExposeIntermediateTensorsPatternList
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.util.test import download_model, test_model_details

model_details_expint = {
    "FINN-TFC_W2A2": {"n_quant_outputs": 4},
    "FINN-CNV_W2A2": {"n_quant_outputs": 9},
    "MobileNetv1-w4a4": {"n_quant_outputs": 27},
}

# inherit basics for matching testcases from test util
model_details = {k: v for (k, v) in test_model_details.items() if k in model_details_expint.keys()}
model_details = {**model_details, **model_details_expint}


@pytest.mark.parametrize("model_name", model_details.keys())
def test_expose_intermediate(model_name):
    model = download_model(model_name, do_cleanup=True, return_modelwrapper=True)
    # do folding for weights
    model = model.transform(FoldConstants(exclude_op_types=[]))
    # break out all dynamic (non-weight) quantizer outputs
    pattern_list = ["Quant"]
    model = model.transform(ExposeIntermediateTensorsPatternList(pattern_list, dynamic_only=True))
    assert len(model.graph.output) == model_details_expint[model_name]["n_quant_outputs"] + 1
